@@ -11,24 +11,32 @@
 # FOR A PARTICULAR PURPOSE
 #
 ##############################################################################
-"""Handy standard storage machinery
+"""Storage base class that is mostly a mistake
 
-$Id: BaseStorage.py 113732 2010-06-21 15:31:11Z ctheune $
+The base class here is tightly coupled with its subclasses and
+its use is not recommended.  It's still here for historical reasons.
 """
-import cPickle
+from __future__ import print_function
+
 import threading
 import time
 import logging
+import sys
 from struct import pack as _structpack, unpack as _structunpack
 
+import zope.interface
 from persistent.TimeStamp import TimeStamp
 
+import ZODB.interfaces
 from ZODB import POSException
-from ZODB.utils import z64, oid_repr
+from ZODB.utils import z64, oid_repr, byte_ord, byte_chr
 from ZODB.UndoLogCompatible import UndoLogCompatible
+from ZODB._compat import dumps, _protocol, py2_hasattr
 
 log = logging.getLogger("ZODB.BaseStorage")
 
+
+
 class BaseStorage(UndoLogCompatible):
     """Base class that supports storage implementations.
 
@@ -77,16 +85,18 @@ class BaseStorage(UndoLogCompatible):
         log.debug("create storage %s", self.__name__)
 
         # Allocate locks:
-        l = threading.RLock()
-        self._lock_acquire = l.acquire
-        self._lock_release = l.release
-        l = threading.Lock()
-        self._commit_lock_acquire = l.acquire
-        self._commit_lock_release = l.release
+        self._lock = threading.RLock()
+        self.__commit_lock = threading.Lock()
+
+        # Comment out the following 4 lines to debug locking:
+        self._lock_acquire = self._lock.acquire
+        self._lock_release = self._lock.release
+        self._commit_lock_acquire = self.__commit_lock.acquire
+        self._commit_lock_release = self.__commit_lock.release
 
         t = time.time()
         t = self._ts = TimeStamp(*(time.gmtime(t)[:5] + (t%60,)))
-        self._tid = repr(t)
+        self._tid = t.raw()
 
         # ._oid is the highest oid in use (0 is always in use -- it's
         # a reserved oid for the root object).  Our new_oid() method
@@ -98,6 +108,45 @@ class BaseStorage(UndoLogCompatible):
         else:
             self._oid = oid
 
+    ########################################################################
+    # The following methods are normally overridden on instances,
+    # except when debugging:
+
+    def _lock_acquire(self, *args):
+        f = sys._getframe(1)
+        sys.stdout.write("[la(%s:%s)\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+        self._lock.acquire(*args)
+        sys.stdout.write("la(%s:%s)]\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+
+    def _lock_release(self, *args):
+        f = sys._getframe(1)
+        sys.stdout.write("[lr(%s:%s)\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+        self._lock.release(*args)
+        sys.stdout.write("lr(%s:%s)]\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+
+    def _commit_lock_acquire(self, *args):
+        f = sys._getframe(1)
+        sys.stdout.write("[ca(%s:%s)\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+        self.__commit_lock.acquire(*args)
+        sys.stdout.write("ca(%s:%s)]\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+
+    def _commit_lock_release(self, *args):
+        f = sys._getframe(1)
+        sys.stdout.write("[cr(%s:%s)\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+        self.__commit_lock.release(*args)
+        sys.stdout.write("cr(%s:%s)]\n" % (f.f_code.co_filename, f.f_lineno))
+        sys.stdout.flush()
+
+    #
+    ########################################################################
+
     def sortKey(self):
         """Return a string that can be used to sort storage instances.
 
@@ -113,7 +162,7 @@ class BaseStorage(UndoLogCompatible):
     def getSize(self):
         return len(self)*300 # WAG!
 
-    def history(self, oid, version='', length=1, filter=None):
+    def history(self, oid, version, length=1, filter=None):
         return ()
 
     def new_oid(self):
@@ -122,9 +171,9 @@ class BaseStorage(UndoLogCompatible):
         self._lock_acquire()
         try:
             last = self._oid
-            d = ord(last[-1])
+            d = byte_ord(last[-1])
             if d < 255:  # fast path for the usual case
-                last = last[:-1] + chr(d+1)
+                last = last[:-1] + byte_chr(d+1)
             else:        # there's a carry out of the last byte
                 last_as_long, = _structunpack(">Q", last)
                 last = _structpack(">Q", last_as_long + 1)
@@ -174,7 +223,8 @@ class BaseStorage(UndoLogCompatible):
         self._lock_acquire()
         try:
             if self._transaction is transaction:
-                return
+                raise POSException.StorageTransactionError(
+                    "Duplicate tpc_begin calls for same transaction")
             self._lock_release()
             self._commit_lock_acquire()
             self._lock_acquire()
@@ -185,16 +235,17 @@ class BaseStorage(UndoLogCompatible):
             desc = transaction.description
             ext = transaction._extension
             if ext:
-                ext = cPickle.dumps(ext, 1)
+                ext = dumps(ext, _protocol)
             else:
                 ext = ""
+
             self._ude = user, desc, ext
 
             if tid is None:
                 now = time.time()
                 t = TimeStamp(*(time.gmtime(now)[:5] + (now % 60,)))
                 self._ts = t = t.laterThan(self._ts)
-                self._tid = repr(t)
+                self._tid = t.raw()
             else:
                 self._ts = TimeStamp(tid)
                 self._tid = tid
@@ -216,7 +267,8 @@ class BaseStorage(UndoLogCompatible):
         self._lock_acquire()
         try:
             if transaction is not self._transaction:
-                return
+                raise POSException.StorageTransactionError(
+                    "tpc_vote called with wrong transaction")
             self._vote()
         finally:
             self._lock_release()
@@ -236,14 +288,14 @@ class BaseStorage(UndoLogCompatible):
         self._lock_acquire()
         try:
             if transaction is not self._transaction:
-                return
+                raise POSException.StorageTransactionError(
+                    "tpc_finish called with wrong transaction")
             try:
                 if f is not None:
                     f(self._tid)
                 u, d, e = self._ude
                 self._finish(self._tid, u, d, e)
                 self._clear_temp()
-                return self._tid
             finally:
                 self._ude = None
                 self._transaction = None
@@ -256,6 +308,10 @@ class BaseStorage(UndoLogCompatible):
         """
         pass
 
+    def lastTransaction(self):
+        with self._lock:
+            return self._ltid
+
     def getTid(self, oid):
         self._lock_acquire()
         try:
@@ -310,7 +366,7 @@ def copy(source, dest, verbose=0):
     # using store().  However, if we use store, then
     # copyTransactionsFrom() may fail with VersionLockError or
     # ConflictError.
-    restoring = hasattr(dest, 'restore')
+    restoring = py2_hasattr(dest, 'restore')
     fiter = source.iterator()
     for transaction in fiter:
         tid = transaction.tid
@@ -319,24 +375,24 @@ def copy(source, dest, verbose=0):
         else:
             t = TimeStamp(tid)
             if t <= _ts:
-                if ok: print ('Time stamps out of order %s, %s' % (_ts, t))
+                if ok: print(('Time stamps out of order %s, %s' % (_ts, t)))
                 ok = 0
                 _ts = t.laterThan(_ts)
-                tid = `_ts`
+                tid = _ts.raw()
             else:
                 _ts = t
                 if not ok:
-                    print ('Time stamps back in order %s' % (t))
+                    print(('Time stamps back in order %s' % (t)))
                     ok = 1
 
         if verbose:
-            print _ts
+            print(_ts)
 
         dest.tpc_begin(transaction, tid, transaction.status)
         for r in transaction:
             oid = r.oid
             if verbose:
-                print oid_repr(oid), r.version, len(r.data)
+                print(oid_repr(oid), r.version, len(r.data))
             if restoring:
                 dest.restore(oid, r.tid, r.data, r.version,
                              r.data_txn, transaction)
@@ -348,10 +404,50 @@ def copy(source, dest, verbose=0):
         dest.tpc_vote(transaction)
         dest.tpc_finish(transaction)
 
-    fiter.close()
 
-class TransactionRecord:
+# defined outside of BaseStorage to facilitate independent reuse.
+# just depends on _transaction attr and getTid method.
+def checkCurrentSerialInTransaction(self, oid, serial, transaction):
+    if transaction is not self._transaction:
+        raise POSException.StorageTransactionError(self, transaction)
+
+    committed_tid = self.getTid(oid)
+    if committed_tid != serial:
+        raise POSException.ReadConflictError(
+            oid=oid, serials=(committed_tid, serial))
+
+BaseStorage.checkCurrentSerialInTransaction = checkCurrentSerialInTransaction
+
+@zope.interface.implementer(ZODB.interfaces.IStorageTransactionInformation)
+class TransactionRecord(object):
     """Abstract base class for iterator protocol"""
 
-class DataRecord:
+
+    def __init__(self, tid, status, user, description, extension):
+        self.tid = tid
+        self.status = status
+        self.user = user
+        self.description = description
+        self.extension = extension
+
+    # XXX This is a workaround to make the TransactionRecord compatible with a
+    # transaction object because it is passed to tpc_begin().
+    def _ext_set(self, value):
+        self.extension = value
+    def _ext_get(self):
+        return self.extension
+    _extension = property(fset=_ext_set, fget=_ext_get)
+
+
+@zope.interface.implementer(ZODB.interfaces.IStorageRecordInformation)
+class DataRecord(object):
     """Abstract base class for iterator protocol"""
+
+
+    version = ''
+
+    def __init__(self, oid, tid, data, prev):
+        self.oid = oid
+        self.tid = tid
+        self.data = data
+        self.data_txn = prev
