@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 120291 2011-02-11 23:50:41Z jim $"""
+$Id$"""
 
 import logging
 import sys
@@ -34,7 +34,7 @@ from ZODB.blob import Blob, rename_or_copy_blob, remove_committed_dir
 from transaction.interfaces import ISavepointDataManager
 from transaction.interfaces import IDataManagerSavepoint
 from transaction.interfaces import ISynchronizer
-from zope.interface import implements
+from zope.interface import implementer
 
 import transaction
 
@@ -50,9 +50,11 @@ from ZODB.POSException import POSKeyError
 from ZODB.serialize import ObjectWriter, ObjectReader
 from ZODB.utils import p64, u64, z64, oid_repr, positive_id
 from ZODB import utils
+import six
 
 global_reset_counter = 0
 
+
 def resetCaches():
     """Causes all connection caches to be reset as connections are reopened.
 
@@ -66,13 +68,19 @@ def resetCaches():
     global global_reset_counter
     global_reset_counter += 1
 
+
+def className(obj):
+    cls = type(obj)
+    return "%s.%s" % (cls.__module__, cls.__name__)
+
+
+@implementer(IConnection,
+             ISavepointDataManager,
+             IPersistentDataManager,
+             ISynchronizer)
 class Connection(ExportImport, object):
     """Connection to ZODB for loading and storing objects."""
 
-    implements(IConnection,
-               ISavepointDataManager,
-               IPersistentDataManager,
-               ISynchronizer)
 
 
     _code_timestamp = 0
@@ -260,14 +268,14 @@ class Connection(ExportImport, object):
     def cacheMinimize(self):
         """Deactivate all unmodified objects in the cache.
         """
-        for connection in self.connections.itervalues():
+        for connection in six.itervalues(self.connections):
             connection._cache.minimize()
 
     # TODO: we should test what happens when cacheGC is called mid-transaction.
     def cacheGC(self):
         """Reduce cache size to target size.
         """
-        for connection in self.connections.itervalues():
+        for connection in six.itervalues(self.connections):
             connection._cache.incrgc()
 
     __onCloseCallbacks = None
@@ -321,6 +329,10 @@ class Connection(ExportImport, object):
                 # get back here.
         else:
             self.opened = None
+            
+        am = self._db._activity_monitor
+        if am is not None:
+            am.closedConnection(self)        
 
     def db(self):
         """Returns a handle to the database this connection belongs to."""
@@ -341,7 +353,7 @@ class Connection(ExportImport, object):
         try:
             if self._txn_time is None:
                 self._txn_time = tid
-            elif (tid < self._txn_time) and (tid is not None):
+            elif (tid is not None) and (tid < self._txn_time):
                 raise AssertionError("invalidations out of order, %r < %r"
                                      % (tid, self._txn_time))
 
@@ -489,7 +501,7 @@ class Connection(ExportImport, object):
             if invalidated is None:
                 # special value: the transaction is so old that
                 # we need to flush the whole cache.
-                self._cache.invalidate(self._cache.cache_data.keys())
+                self._cache.invalidate(list(self._cache.cache_data.keys()))
             elif invalidated:
                 self._cache.invalidate(invalidated)
 
@@ -562,7 +574,7 @@ class Connection(ExportImport, object):
         else:
             self._commit(transaction)
 
-        for oid, serial in self._readCurrent.iteritems():
+        for oid, serial in six.iteritems(self._readCurrent):
             try:
                 self._storage.checkCurrentSerialInTransaction(
                     oid, serial, transaction)
@@ -704,7 +716,7 @@ class Connection(ExportImport, object):
 
         if not serial:
             return
-        if not isinstance(serial, str):
+        if not isinstance(serial, bytes):
             raise serial
         obj = self._cache.get(oid, None)
         if obj is None:
@@ -775,7 +787,7 @@ class Connection(ExportImport, object):
             return
         try:
             s = vote(transaction)
-        except ReadConflictError, v:
+        except ReadConflictError as v:
             if v.oid:
                 self._cache.invalidate(v.oid)
             raise
@@ -847,18 +859,23 @@ class Connection(ExportImport, object):
         oid = obj._p_oid
 
         if self.opened is None:
-            msg = ("Shouldn't load state for %s "
-                   "when the connection is closed" % oid_repr(oid))
-            self._log.error(msg)
-            raise ConnectionStateError(msg)
+            msg = ("Shouldn't load state for %s %s "
+                   "when the connection is closed"
+                   % (className(obj), oid_repr(oid)))
+            try:
+                raise ConnectionStateError(msg)
+            except:
+                self._log.exception(msg)
+                raise
 
         try:
             self._setstate(obj)
         except ConflictError:
             raise
         except:
-            self._log.error("Couldn't load state for %s", oid_repr(oid),
-                            exc_info=sys.exc_info())
+            self._log.exception("Couldn't load state for %s %s",
+                                className(obj), oid_repr(oid),
+                                exc_info=sys.exc_info())
             raise
 
     def _setstate(self, obj):
@@ -991,7 +1008,8 @@ class Connection(ExportImport, object):
     def readCurrent(self, ob):
         assert ob._p_jar is self
         assert ob._p_oid is not None and ob._p_serial is not None
-        self._readCurrent[ob._p_oid] = ob._p_serial
+        if ob._p_serial != z64:
+            self._readCurrent[ob._p_oid] = ob._p_serial
 
     # persistent.interfaces.IPersistentDatamanager
     ##########################################################################
@@ -1008,7 +1026,7 @@ class Connection(ExportImport, object):
         for k,v in items:
             del everything[k]
         # return a list of [ghosts....not recently used.....recently used]
-        return everything.items() + items
+        return list(everything.items()) + items
 
     def open(self, transaction_manager=None, delegate=True):
         """Register odb, the DB that this Connection uses.
@@ -1066,7 +1084,7 @@ class Connection(ExportImport, object):
             self._reader._cache = cache
 
     def _release_resources(self):
-        for c in self.connections.itervalues():
+        for c in six.itervalues(self.connections):
             if c._mvcc_storage:
                 c._storage.release()
             c._storage = c._normal_storage = None
@@ -1153,7 +1171,7 @@ class Connection(ExportImport, object):
         self._savepoint_storage = None
         try:
             self._log.debug("Committing savepoints of size %s", src.getSize())
-            oids = src.index.keys()
+            oids = sorted(src.index.keys())
 
             # Copy invalidating and creating info from temporary storage:
             self._modified.extend(oids)
@@ -1217,9 +1235,9 @@ class Connection(ExportImport, object):
     # Savepoint support
     #####################################################################
 
+@implementer(IDataManagerSavepoint)
 class Savepoint:
 
-    implements(IDataManagerSavepoint)
 
     def __init__(self, datamanager, state):
         self.datamanager = datamanager
@@ -1228,10 +1246,10 @@ class Savepoint:
     def rollback(self):
         self.datamanager._rollback(self.state)
 
+@implementer(IBlobStorage)
 class TmpStore:
     """A storage-like thing to support savepoints."""
 
-    implements(IBlobStorage)
 
     def __init__(self, storage):
         self._storage = storage
@@ -1241,10 +1259,10 @@ class TmpStore:
             ):
             setattr(self, method, getattr(storage, method))
 
-        self._file = tempfile.TemporaryFile()
+        self._file = tempfile.TemporaryFile(prefix='TmpStore')
         # position: current file position
         # _tpos: file position at last commit point
-        self.position = 0L
+        self.position = 0
         # index: map oid to pos of last committed version
         self.index = {}
         self.creating = {}
@@ -1297,7 +1315,7 @@ class TmpStore:
 
         targetpath = self._getBlobPath()
         if not os.path.exists(targetpath):
-            os.makedirs(targetpath, 0700)
+            os.makedirs(targetpath, 0o700)
 
         targetname = self._getCleanFilename(oid, serial)
         rename_or_copy_blob(blobfilename, targetname, chmod=False)
