@@ -1,5 +1,4 @@
-#!/usr/bin/env python2.3
-
+#!/usr/bin/env python
 ##############################################################################
 #
 # Copyright (c) 2001, 2002 Zope Foundation and Contributors.
@@ -14,30 +13,34 @@
 #
 ##############################################################################
 """Tool to dump the last few transactions from a FileStorage."""
-
+from __future__ import print_function
 from ZODB.fstools import prev_txn
 
 import binascii
 import getopt
-import sha
 import sys
 
+try:
+    from hashlib import sha1
+except ImportError:
+    from sha import sha as sha1
+
 def main(path, ntxn):
-    f = open(path, "rb")
-    f.seek(0, 2)
-    th = prev_txn(f)
-    i = ntxn
-    while th and i > 0:
-        hash = sha.sha(th.get_raw_data()).digest()
-        l = len(str(th.get_timestamp())) + 1
-        th.read_meta()
-        print "%s: hash=%s" % (th.get_timestamp(),
-                               binascii.hexlify(hash))
-        print ("user=%r description=%r length=%d"
-               % (th.user, th.descr, th.length))
-        print
-        th = th.prev_txn()
-        i -= 1
+    with open(path, "rb") as f:
+        f.seek(0, 2)
+        th = prev_txn(f)
+        i = ntxn
+        while th and i > 0:
+            hash = sha1(th.get_raw_data()).digest()
+            l = len(str(th.get_timestamp())) + 1
+            th.read_meta()
+            print("%s: hash=%s" % (th.get_timestamp(),
+                                   binascii.hexlify(hash).decode()))
+            print(("user=%r description=%r length=%d offset=%d"
+                   % (th.user, th.descr, th.length, th.get_data_offset())))
+            print()
+            th = th.prev_txn()
+            i -= 1
 
 def Main():
     ntxn = 10
@@ -47,7 +50,6 @@ def Main():
         if k == '-n':
             ntxn = int(v)
     main(path, ntxn)
-    
 
 if __name__ == "__main__":
     Main()
